"""
Polyfill for a JSON field in SQLite.
Newer versions of sqlite (>= 3.9.0) have a native JSON extension, which we use.
If the sqlite version is lower, we roll a less-optimal replacement.
"""
from typing import Dict, cast

import peewee as pw
from peewee import sqlite3
import json

if sqlite3.sqlite_version_info < (3, 9, 0):  # type: ignore

    class JSONField(pw.CharField):  # pragma: no cover
        """
        Polyfill class to provide a JSON field
        """

        def db_value(self, value: Dict) -> str:
            """
            Convert a value to a string for storage in a `CharField`
            :param value: The value to store
            :return: The JSON string
            """
            return json.dumps(value)

        def python_value(self, value: str) -> Dict:
            """
            Convert a string value from the database back to what it was.
            :param value: The string value
            :return: Parsed JSON value
            """
            return cast(Dict, json.loads(value))


else:  # pragma: no cover
    # flake8: noqa
    from playhouse.sqlite_ext import JSONField  # type: ignore



import shutil

import click
from typing import List, Tuple, Optional

from .util import shorten


def _do_align(s: str, align: str, width: int, padstr: str) -> str:
    length = len(click.unstyle(s))
    if length > width:
        s = shorten(s, width)

    if align == "l":
        return s + (width - length) * padstr
    elif align == "r":
        return (width - length) * padstr + s
    elif align == "c":
        left = (width - length) // 2
        right = width - length - left
        return padstr * left + s + padstr * right
    else:
        raise ValueError(f"Invalid align string '{align}'")


def _format_row(
    cols: Tuple[str, ...],
    widths: List[int],
    aligns: Tuple[str, ...],
    max_width: Optional[int] = None,
    padstr: str = " ",
) -> str:
    assert len(cols) == len(aligns), "Number of aligns must match columns"
    colstrs = ["" for _ in cols]
    stretch_col = None
    for idx, (col, width, align) in enumerate(zip(cols, widths, aligns)):
        if align.endswith("+"):
            stretch_col = idx
            colstr = _do_align(col, align[:-1], width, padstr)
        else:
            colstr = _do_align(col, align, width, padstr)

        colstrs[idx] = colstr

    if stretch_col is not None and max_width is not None:
        total_len = sum(
            len(click.unstyle(c)) for i, c in enumerate(colstrs) if i != stretch_col
        ) + (len(cols) - 1) * len(padstr)
        align = aligns[stretch_col][:-1]
        width = max_width - total_len

        col = cols[stretch_col]

        colstr = _do_align(col, align, width, padstr)

        colstrs[stretch_col] = colstr

    return " ".join(colstrs)


def format_table(
    headers: Tuple[str, ...],
    rows: List[Tuple[str, ...]],
    align: Tuple[str, ...],
    width: Optional[int] = None,
) -> str:
    assert len(headers) == len(align), "Number of aligns must match columns"
    if width is None and any("+" in a for a in align):
        width, _ = shutil.get_terminal_size((80, 40))

    col_widths = [len(click.unstyle(h)) for h in headers]
    for row in rows:
        for idx, col in enumerate(row):
            col_widths[idx] = max(col_widths[idx], len(click.unstyle(col)))

    output = ""
    output += _format_row(headers, col_widths, align, max_width=width) + "\n"

    output += (
        _format_row(
            tuple(["" for _ in col_widths]),
            col_widths,
            align,
            max_width=width,
            padstr="-",
        )
        + "\n"
    )

    for row in rows:
        output += _format_row(row, col_widths, align, max_width=width) + "\n"

    return output[:-1]



import click
import pytest
from click import unstyle

from kong.table import format_table, _do_align


def test_no_stretch():
    headers = ["alpha", "beta", "gamma"]
    rows = [
        ["delta", "omega", "psi"],
        ["echo", "charlie", "bingo"],
        ["nebukadnezar", "otto", "kilo"],
    ]
    align = ["r", "c", "l"]

    s = format_table(headers, rows, align, width=None)
    print(s)

    assert (
        s
        == """
       alpha  beta   gamma
------------ ------- -----
       delta  omega  psi  
        echo charlie bingo
nebukadnezar  otto   kilo 
"""[
            1:-1
        ]
    )

    s = format_table(headers, rows, align, width=80)
    print(s)

    assert (
        s
        == """
       alpha  beta   gamma
------------ ------- -----
       delta  omega  psi  
        echo charlie bingo
nebukadnezar  otto   kilo 
"""[
            1:-1
        ]
    )


def test_stretch():
    headers = ["alpha", "beta", "gamma"]
    rows = [
        ["delta", "omega", "psi"],
        ["echo", "charlie", "bingo"],
        ["nebukadnezar", "otto", "kilo"],
    ]
    align = ["r+", "c", "l"]
    s = format_table(headers, rows, align, width=80)
    assert (
        s
        == """
                                                             alpha  beta   gamma
------------------------------------------------------------------ ------- -----
                                                             delta  omega  psi  
                                                              echo charlie bingo
                                                      nebukadnezar  otto   kilo 
"""[
            1:-1
        ]
    )

    for line in s.split("\n"):
        assert len(line) == 80

    align = ["l", "r+", "l"]
    s = format_table(headers, rows, align, width=100)
    assert (
        s
        == """
alpha                                                                                     beta gamma
------------ --------------------------------------------------------------------------------- -----
delta                                                                                    omega psi  
echo                                                                                   charlie bingo
nebukadnezar                                                                              otto kilo 
"""[
            1:-1
        ]
    )

    for line in s.split("\n"):
        assert len(line) == 100

    align = ["r+", "r", "l"]
    s = format_table(headers, rows, align, width=100)
    print()
    print(s)
    assert (
        s
        == """
                                                                                 alpha    beta gamma
-------------------------------------------------------------------------------------- ------- -----
                                                                                 delta   omega psi  
                                                                                  echo charlie bingo
                                                                          nebukadnezar    otto kilo 
"""[
            1:-1
        ]
    )

    for line in s.split("\n"):
        assert len(line) == 100

    rows = [
        ["delta", "omega", "psi"],
        ["echo", "charlie tango tina purple rain", "bingo"],
        ["nebukadnezar", "otto", "kilo"],
    ]
    align = ["l", "c+", "l"]
    s = format_table(headers, rows, align, width=40)
    print(s)

    assert (
        s
        == """
alpha                beta          gamma
------------ --------------------- -----
delta                omega         psi  
echo         charlie t...rple rain bingo
nebukadnezar         otto          kilo 
"""[
            1:-1
        ]
    )

    for line in s.split("\n"):
        assert len(line) == 40


def test_colors():
    headers = ["alpha", "beta", click.style("gamma", bg="green")]
    rows = [
        ["delta", "omega", "psi"],
        ["echo", click.style("charlie", fg="red"), "bingo"],
        [click.style("nebukadnezar", bold=True), "otto", "kilo"],
    ]
    align = ["r", "c", "l"]

    s = format_table(headers, rows, align, width=None)
    print()
    print(s)

    assert (
        unstyle(s)
        == """
       alpha  beta   gamma
------------ ------- -----
       delta  omega  psi  
        echo charlie bingo
nebukadnezar  otto   kilo 
"""[
            1:-1
        ]
    )

    align = ["r+", "c", "l"]

    s = format_table(headers, rows, align, width=100)
    print()
    print(s)

    assert (
        unstyle(s)
        == """
                                                                                 alpha  beta   gamma
-------------------------------------------------------------------------------------- ------- -----
                                                                                 delta  omega  psi  
                                                                                  echo charlie bingo
                                                                          nebukadnezar  otto   kilo 
"""[
            1:-1
        ]
    )

def test_stretch_shorten():
    headers = ["alpha", "beta", click.style("gamma", bg="green")]
    rows = [
        ["delta", "omega", "psi"],
        ["echo", click.style("charlie", fg="red"), "bingo"],
        [click.style("nebukadnezar prometheus apollo jupiter", bold=True), "otto", "kilo"],
    ]
    align = ["r+", "c", "l"]
    s = format_table(headers, rows, align, width=50)
    print()
    print(s)

    assert (
        unstyle(s)
        == """
                               alpha  beta   gamma
------------------------------------ ------- -----
                               delta  omega  psi  
                                echo charlie bingo
nebukadnezar pro...us apollo jupiter  otto   kilo 
"""[
           1:-1
           ]
    )


def test_do_align():
    s = "abcabc"
    assert "abcabc    " == _do_align(s, "l", 10, " ")
    assert "    abcabc" == _do_align(s, "r", 10, " ")
    assert "  abcabc  " == _do_align(s, "c", 10, " ")
    assert " abcxabc  " == _do_align("abcxabc", "c", 10, " ")
    with pytest.raises(ValueError):
        _do_align("abcabc", "k", 10, " ")

    s = "long long long long"
    assert 10 == len(_do_align(s, "r", 10, " "))
    s = click.style("nebukadnezar prometheus apollo jupiter", bold=True)
    out = _do_align(s, "r", 36, " ")
    assert len(click.unstyle(out)) == 36
    assert click.unstyle(out) == "nebukadnezar pro...us apollo jupiter"
    assert click.unstyle(out) != out

    s = click.style("abcabc", fg="red")
    assert _do_align(s, "l", 10, " ") == f"{s}    "
    assert _do_align(s, "r", 10, " ") == f"    {s}"
    assert _do_align(s, "c", 10, " ") == f"  {s}  "
    s = click.style("abcxabc", fg="red")
    assert _do_align(s, "c", 10, " ") == f" {s}  "




import copy
import datetime
import shlex
import cmd
import readline
import os
import subprocess
import sys
import time
from concurrent.futures import wait, ThreadPoolExecutor

import humanfriendly
import sh
from typing import Any, Callable, List, Optional, Union, Iterable, cast, Dict, Tuple
import shutil

import click
import peewee as pw
from click import style
from kong.model.job import color_dict
from .table import format_table

from .util import shorten_path, Spinner
from .state import DoesNotExist
from .config import APP_NAME, APP_DIR
from .logger import logger
from .model.job import Job
from .model.folder import Folder
from . import state

history_file = os.path.join(APP_DIR, "history")


def complete_path(cwd: Folder, path: str) -> List[str]:
    logger.debug("Completion of '%s'", path)
    if path.endswith("/"):
        folder = Folder.find_by_path(path, cwd)
        prefix = ""
    else:
        head, prefix = os.path.split(path)
        folder = Folder.find_by_path(head, cwd)

    assert folder is not None
    options = []
    for child in folder.children:
        if child.name.startswith(prefix):
            options.append(child.name + "/")
    return options


def parse_arguments(fn: Any) -> Callable[[Any, str], None]:
    _, prog_name = fn.__name__.split("_", 1)

    orig_fn = copy.deepcopy(fn)
    fn = click.pass_obj(fn)
    command = click.command()(fn)

    def wrapped(self: Any, argstr: str) -> None:
        argv = shlex.split(argstr)
        logger.debug("%s", argv)
        try:
            command.main(
                args=argv,
                prog_name=prog_name,
                standalone_mode=False,
                obj=self,
                help_option_names=["-h", "--help"],
            )
        except click.MissingParameter:
            click.echo(f"usage: {fn.__doc__}")

    wrapped.__doc__ = fn.__doc__  # type: ignore
    wrapped.__name__ = fn.__name__  # type: ignore
    wrapped.__orig_fn__ = orig_fn  # type: ignore

    return wrapped


class Repl(cmd.Cmd):
    intro = f"This is {APP_NAME} shell"
    prompt = f"({APP_NAME} > /) "

    _raise: bool = False

    def __init__(self, state: state.State) -> None:
        self.state = state
        super().__init__()

    def precmd(self, line: str) -> str:
        if line != "":
            logger.debug("called '%s'", line)
        return line

    # For whatever reason, Cmd.completedefault has this signature...
    def completedefault(self, *args: Any) -> List[str]:
        # forward to typed implementation
        return self.completedefault_impl(*args)

    def completedefault_impl(
        self, text: str, line: str, begidx: int, endidx: int
    ) -> List[str]:
        logger.debug(
            "text: %s, line: %s, begidx: %d, endidx: %d", text, line, begidx, endidx,
        )

        # find component
        parts = shlex.split(line)
        base: Optional[str] = None
        for i, part in enumerate(parts):  # pragma: no branch
            prelength = len(" ".join(parts[: i + 1]))
            if prelength >= begidx:
                base = part
                break
        assert base is not None, "Error extracting active part"

        return complete_path(self.state.cwd, base)  # type: ignore

    def onecmd(self, *args: str) -> bool:
        try:
            res = super().onecmd(*args)
            # logger.debug(
            #     "Writing history of length %d to file %s",
            #     self.state.config.history_length,
            #     history_file,
            # )
            readline.set_history_length(self.state.config.history_length)
            readline.write_history_file(history_file)
            return res
        except (BaseException, Exception) as e:
            logger.debug("Exception occured", exc_info=True)
            click.secho(f"{e}", fg="red")
            if self._raise:
                raise e
        return False

    @parse_arguments
    @click.argument("dir", default="", required=False)
    @click.option(
        "--refresh",
        "-R",
        is_flag=True,
        help="Synchronize status on all jobs selected for display",
    )
    @click.option(
        "--recursive",
        "-r",
        is_flag=True,
        help="Recursively select all jobs from target directory",
    )
    @click.option(
        "--show-sizes",
        "-s",
        is_flag=True,
        help="Collect size of job outputs. Note: this can potentially take a while.",
    )
    @click.option(
        "--status",
        "-S",
        "status_filter_str",
        type=click.Choice([s.name for s in Job.Status]),
        help="Only list jobs with this status. Will still show other jobs in folder summaries",
    )
    def do_ls(
        self,
        dir: str,
        refresh: bool,
        recursive: bool,
        show_sizes: bool,
        status_filter_str: Optional[str],
    ) -> None:
        "List the directory content of DIR: jobs and folders"
        try:
            ex: Optional[ThreadPoolExecutor] = None
            if show_sizes:
                ex = ThreadPoolExecutor()

            folders, jobs = self.state.ls(dir, refresh=refresh)

            if recursive:
                arg_folder = Folder.find_by_path(dir, self.state.cwd)
                assert arg_folder is not None  # should be a folder
                jobs = list(arg_folder.jobs_recursive())

            with Spinner("Refreshing jobs", persist=False, enabled=refresh):

                if refresh:
                    jobs = cast(list, self.state.refresh_jobs(jobs))

            def get_size(job: Job) -> int:
                return job.size(cast(ThreadPoolExecutor, ex))

            def get_folder_size(folder: Folder) -> int:
                # print("get folder size: ", folder.path)
                return sum(
                    cast(ThreadPoolExecutor, ex).map(get_size, folder.jobs_recursive())
                )

            folder_sizes: List[int] = []
            jobs_sizes: List[int] = []
            if show_sizes:
                ex_ = cast(ThreadPoolExecutor, ex)
                with Spinner("Calculating output sizes", persist=False):
                    folder_size_futures = []
                    for folder in folders:
                        folder_size_futures.append(ex_.submit(get_folder_size, folder))

                    job_size_futures = []
                    for job in jobs:
                        job_size_futures.append(ex_.submit(get_size, job))

                    wait(folder_size_futures)
                    folder_sizes = [f.result() for f in folder_size_futures]
                    wait(job_size_futures)
                    jobs_sizes = [f.result() for f in job_size_futures]

            if len(folders) > 0:
                with Spinner("Collection folder information", persist=False):
                    headers = ["name"]
                    align = ["l+"]

                    if show_sizes:
                        headers.append("output size")
                        align = ["l", "l+"]

                    for s in Job.Status:
                        headers.append(click.style(s.name, fg=color_dict[s]))
                        align.append("r")

                rows = []
                for idx, folder in enumerate(folders):
                    counts = folder.job_stats()

                    output = ""
                    for k, c in counts.items():
                        output += style(f" {c:> 6d}{k.name[:1]}", fg=color_dict[k])

                    row = [folder.name]
                    if show_sizes:
                        row.append(humanfriendly.format_size(folder_sizes[idx]))
                    # row += [output]
                    for k, c in counts.items():
                        row.append(click.style(str(c), fg=color_dict[k]))

                    rows.append(tuple(row))

                click.echo(format_table(tuple(headers), rows, align=tuple(align)))

            if len(folders) > 0 and len(jobs) > 0:
                print()

            if len(jobs) > 0:
                with Spinner("Collection job information", persist=False):
                    headers = ["job id"]
                    align = ["l"]

                    if show_sizes:
                        headers.append("output size")
                        align.append("l")

                    headers += ["batch job id", "created", "updated", "status"]
                    align += ["r+", "l", "l", "l"]

                    def dtfmt(dt: datetime.datetime) -> str:
                        return dt.strftime("%Y-%m-%d %H:%M:%S")

                    rows = []
                    status_filter = (
                        Job.Status[status_filter_str]
                        if status_filter_str is not None
                        else None
                    )
                    for idx, job in enumerate(jobs):

                        if status_filter is not None:
                            if job.status != status_filter:
                                continue

                        job_id = str(job.job_id)
                        batch_job_id = str(job.batch_job_id)
                        _, status_name = str(job.status).split(".", 1)
                        color = color_dict[job.status]
                        row = [job_id]
                        if show_sizes:
                            row.append(humanfriendly.format_size(jobs_sizes[idx]))

                        row += [
                            batch_job_id,
                            dtfmt(job.created_at),
                            dtfmt(job.updated_at),
                            status_name,
                        ]

                        rows.append(tuple(click.style(c, fg=color) for c in row))

                click.echo(format_table(tuple(headers), rows, align=tuple(align)))

                if show_sizes:
                    click.echo(
                        "Size of jobs listed above: "
                        + click.style(
                            humanfriendly.format_size(sum(jobs_sizes)), bold=True
                        )
                    )

        except pw.DoesNotExist:
            click.secho(f"Folder {dir} does not exist", fg="red")
        finally:
            if ex is not None:
                ex.shutdown()

    @parse_arguments
    @click.argument("path")
    @click.option(
        "--create-parent",
        "-p",
        is_flag=True,
        help="Create parent directories to 'path' if they don't exist",
    )
    def do_mkdir(self, path: str, create_parent: bool) -> None:
        "Create a directory at PATH"
        try:
            self.state.mkdir(path, create_parent=create_parent)
        except state.CannotCreateError:
            click.secho(f"Cannot create folder at '{path}'", fg="red")
        except pw.IntegrityError:
            click.secho(
                f"Folder {path} in {self.state.cwd.path} already exists", fg="red"
            )

    @parse_arguments
    @click.argument("path", required=False, default="")
    def do_cd(self, path: str = "") -> None:
        """
        Change current working directory into PATH.

        \b
        Examples:
            cd # root /
            cd . # current directory
            cd .. # parent directory
            cd another/folder # some folder from this directory
            cd ../another # path but starting from parent
        """
        # find the folder
        try:
            self.state.cd(path)
        except pw.DoesNotExist:
            click.secho(f"Folder {path} does not exist", fg="red")
        self.prompt = f"({APP_NAME} > {shorten_path(self.state.cwd.path, 40)}) "

    @parse_arguments
    @click.argument("src")
    @click.argument("dest")
    def do_mv(self, src: str, dest: str) -> None:
        """
        Move SRC into DEST.

        SRC can be either jobs or folders. If there are jobs in SRC, then
        DEST needs to be a folder that exists. If SRC is exactly one folder, and DEST
        does not exist, SRC will be renamed to DEST. If SRC is more than one folder,
        DEST needs to exist, and SRC folders will be moved into DEST.
        """
        items: List[Union[Job, Folder]] = self.state.mv(src, dest)
        names = []
        for item in items:
            if isinstance(item, Job):
                names.append(str(item.job_id))
            else:
                names.append(item.name)

        click.secho(f"Moved -> {dest}")

    @parse_arguments
    @click.argument("path")
    @click.option(
        "--refresh", "-R", is_flag=True, help="Do a refresh on all jobs found"
    )
    @click.option(
        "--recursive",
        "-r",
        is_flag=True,
        help="Go recursively through PATH to find jobs (if PATH is a folder)",
    )
    @click.option("--full", is_flag=True, help="Do not truncate job info if too long")
    def do_info(self, path: str, refresh: bool, recursive: bool, full: bool) -> None:
        """
        Show information on jobs from PATH.

        PATH can be a job id (range) or a folder.
        """
        jobs = self.state.get_jobs(path, recursive)
        if refresh:
            jobs = list(self.state.refresh_jobs(jobs))

        for job in jobs:
            click.echo(job)
            for field in (
                "batch_job_id",
                "driver",
                "folder",
                "command",
                "cores",
                "status",
                "created_at",
                "updated_at",
            ):
                fg: Optional[str] = None
                if field == "status":
                    fg = color_dict[job.status]
                if field == "command" and not full:
                    cmd = job.command
                    if len(cmd) > 500:
                        cmd = cmd[:500] + "..."
                    click.secho(f"{field}: {cmd}", fg=fg)
                    continue

                click.secho(f"{field}: {str(getattr(job, field))}", fg=fg)
            click.echo("data:")
            for k, v in job.data.items():
                click.secho(f"- {k}: {v}")

    @parse_arguments
    @click.argument("path")
    @click.option("--recursive", "-r", is_flag=True, help="Recursively delete")
    def do_rm(self, path: str, recursive: bool) -> None:
        """
        Delete item at PATH.

        If rm deletes jobs, cleanup routines will be run. Use `rm -r` to delete folders.
        """
        try:
            if self.state.rm(
                path, recursive=recursive, confirm=lambda s: click.confirm(s)
            ):
                click.echo(f"{path} is gone")
        except state.CannotRemoveRoot:
            click.secho("Cannot delete root folder", fg="red")
        except DoesNotExist:
            click.secho(f"Folder {path} does not exist", fg="red")

    @parse_arguments
    def do_cwd(self) -> None:
        "Show the current location"
        click.echo(self.state.cwd.path)

    @parse_arguments
    def do_pwd(self) -> None:
        "Show the current location"
        click.echo(self.state.cwd.path)

    @parse_arguments
    @click.argument("command", nargs=-1)
    @click.option(
        "-a",
        "--argument",
        "arguments_raw",
        multiple=True,
        help="Provide extra arguments like `--argument name=value`",
    )
    def do_create_job(self, command: List[str], arguments_raw: List[str]) -> None:
        """
        Create a job with command COMMAND for processing.
        Additional arguments can be provided and are passed to the driver for verification.
        """
        if len(command) == 0:
            click.secho("Please provide a command to run", fg="red")
            return
        command_str = " ".join(command)

        logger.debug("Raw extra arguments: %s", arguments_raw)
        arg_str: Dict[str, str] = dict(
            [cast(Tuple[str, str], s.split("=", 1)) for s in arguments_raw]
        )

        # cast to int for numeric values
        # @TODO: This might need to be smarter at some point
        arguments: Dict[str, Union[str, int]] = {}
        for k, v in arg_str.items():
            if v.isdigit():
                arguments[k] = int(v)
            else:
                arguments[k] = v

        logger.debug("Got extra arguments: %s", arguments)

        job = self.state.create_job(command=command_str, **arguments)
        click.secho(f"Created job {job}")

    @parse_arguments
    @click.argument("path")
    @click.option("--recursive", "-r", is_flag=True, help="Search recursively for jobs")
    def do_submit_job(self, path: str, recursive: bool) -> None:
        """
        Submit job(s) at PATH.
        """
        self.state.submit_job(path, click.confirm, recursive=recursive)

    @parse_arguments
    @click.argument("path")
    @click.option("--recursive", "-r", is_flag=True, help="Search recursively for jobs")
    def do_kill_job(self, path: str, recursive: bool) -> None:
        """Kill job(s) at PATH"""
        self.state.kill_job(path, recursive=recursive, confirm=click.confirm)

    @parse_arguments
    @click.argument("path")
    @click.option("--recursive", "-r", is_flag=True, help="Search recursively for jobs")
    @click.option(
        "--failed", "-F", is_flag=True, help="Only resubmit jobs in status FAILED"
    )
    def do_resubmit_job(self, path: str, recursive: bool, failed: bool) -> None:
        """Resubmit jobs at PATH."""
        self.state.resubmit_job(
            path, click.confirm, recursive=recursive, failed_only=failed
        )

    @parse_arguments
    @click.argument("path", default=".")
    @click.option(
        "--recursive/--no-recursive",
        "-r/-f",
        default=True,
        help="Select jobs recursively",
    )
    def do_update(self, path: str, recursive: bool) -> None:
        """Update the job at PATH."""
        with Spinner("Updating jobs"):
            jobs = self.state.get_jobs(path, recursive=recursive)
            jobs = list(self.state.refresh_jobs(jobs))
        counts = {k: 0 for k in Job.Status}

        click.echo(f"{len(jobs)} job(s) updated")

        if len(jobs) > 1:
            for job in jobs:
                counts[job.status] += 1

            output = ""
            for k, c in counts.items():
                output += style(f" {c:> 6d}{k.name[:1]}", fg=color_dict[k])

            click.echo(output)

    @parse_arguments
    @click.argument("path")
    @click.option(
        "--number-of-lines",
        "-n",
        default=20,
        type=int,
        help="How many lines to print when starting",
    )
    def do_tail(self, path: str, number_of_lines: int) -> None:
        """
        Tail the stdout of the job at PATH.
        Will wait for the creation of the stdout file if it hasn't already been created.
        """
        from sh import tail  # type: ignore

        jobs = self.state.get_jobs(path)
        assert len(jobs) == 1
        job = jobs[0]

        if not os.path.exists(job.data["stdout"]):
            with Spinner(
                text=f"Waiting for job to create stdout file '{job.data['stdout']}'"
            ):
                while not os.path.exists(job.data["stdout"]):
                    time.sleep(1)
                logger.info("Outfile exists now")

        width, _ = shutil.get_terminal_size((80, 40))
        hw = width // 2
        click.echo("=" * hw + " STDOUT " + "=" * (width - hw - 8))

        def show(line: str) -> None:  # pragma: no cover
            sys.stdout.write(line)

        try:
            proc = tail(
                "-n",
                number_of_lines,
                "-f",
                job.data["stdout"],
                _bg=True,
                _bg_exc=False,
                _out=show,
            )
            try:
                proc.wait()
            except KeyboardInterrupt:  # pragma: no cover
                proc.terminate()
                proc.wait()

        except sh.SignalException_SIGTERM:  # pragma: no cover
            pass

    @parse_arguments
    @click.argument("path")
    def do_less(self, path: str) -> None:
        """Open program less on the stdout file of the job at PATH."""
        jobs = self.state.get_jobs(path)
        assert len(jobs) == 1
        job = jobs[0]

        def reader() -> Iterable[str]:
            with open(job.data["stdout"]) as fp:
                line = fp.readline()
                yield line
                while line:
                    line = fp.readline()
                    yield line

        click.echo_via_pager(reader())

    @parse_arguments
    @click.argument("path", nargs=-1)
    @click.option(
        "--notify/--no-notify",
        default=True,
        help="Send a notification when wait terminates",
    )
    @click.option("--recursive", "-r", is_flag=True, help="Search recursively for jobs")
    @click.option(
        "--poll-interval",
        "-i",
        type=int,
        default=None,
        help="Interval with which to poll for job status updates",
    )
    @click.option(
        "--notify-interval",
        "-n",
        type=str,
        default="30m",
        help="Send periodic status updates with this interval",
    )
    def do_wait(
        self,
        path: List[str],
        notify: bool,
        recursive: bool,
        poll_interval: Optional[int],
        notify_interval: Optional[str],
    ) -> None:
        """
        Wait on the job(s) at PATH.

        This command will periodically poll the driver for updates on the given jobs,
        and will display a tally of the jobs' status.

        Note: notifications will only be sent if a notification provider is configured in
        the config file.
        """

        update_interval: Optional[datetime.timedelta]
        if (
            notify
            and notify_interval is not None
            and notify_interval not in ("none", "None")
        ):
            update_interval = datetime.timedelta(
                seconds=humanfriendly.parse_timespan(notify_interval)
            )
        else:
            update_interval = None

        self.state.wait(
            path,
            notify=notify,
            recursive=recursive,
            poll_interval=poll_interval,
            update_interval=update_interval,
        )

    def do_shell(self, cmd: str) -> None:
        """
        Run the command given in a shell.
        The environment variable $KONG_PWD is set to the current working directory
        """
        logger.debug("cmd: %s", cmd)
        env = os.environ.copy()
        env.update({"KONG_PWD": self.state.cwd.path})
        logger.debug("Expanded: %s", cmd)
        subprocess.run(cmd, shell=True, env=env)

    def do_exit(self, arg: str) -> bool:
        """Exit the repl"""
        return True

    def do_EOF(self, arg: str) -> bool:
        """Helper command to handle Ctrl+D"""
        return self.do_exit(arg)

    def preloop(self) -> None:
        if os.path.exists(history_file):
            # logger.debug("Loading history from %s", history_file)
            readline.read_history_file(history_file)
        else:
            logger.debug("No history file found")

    def postloop(self) -> None:
        pass

    def cmdloop(self, intro: Optional[Any] = None) -> Any:
        print(self.intro)
        while True:
            try:
                super().cmdloop(intro="")
                break
            except KeyboardInterrupt:
                print("^C")

    def emptyline(self) -> bool:
        # do nothing when called with empty
        pass



from datetime import timedelta

from typing import Union


def format_timespan(seconds: Union[float, timedelta]) -> str:
    ...

def parse_timespan(arg: str) -> float:
    ...

def format_size(arg: int) -> str:
    ...



"""
Holds the config schema, a class to load the config, and some helpers to
perform notifications that can be configured in the config file
"""
import os
from datetime import timedelta

import pytimeparse
import typing
import socket
import getpass

import click
import yaml
from schema import And, Optional, Schema, Use
import notifiers  # type: ignore


APP_NAME = "kong"
APP_DIR = click.get_app_dir(APP_NAME, force_posix=True)
CONFIG_FILE = os.path.join(APP_DIR, "config.yml")
DB_FILE = os.path.join(APP_DIR, "database.sqlite")

slurm_schema = Schema(
    {
        Optional("account", default="account"): And(str, len),
        Optional("node_size", default=1): And(int, lambda i: i > 0),
        Optional("default_queue", default="queue"): And(str, len),
        Optional("sacct_delta", default=timedelta(weeks=4)): Use(
            lambda s: timedelta(seconds=pytimeparse.parse(s))
        ),
    }
)

htcondor_schema = Schema(
    {
        Optional("user", default=getpass.getuser()): And(str, len),
        Optional("default_universe", default="vanilla"): And(str, len),
        Optional("submitfile_extra", default=""): str,
    }
)

config_schema = Schema(
    {
        Optional(
            "default_driver", default="kong.drivers.local_driver.LocalDriver"
        ): And(str, len),
        Optional(
            "jobdir", default=lambda: os.path.join(APP_DIR, "jobdir")
        ): os.path.exists,
        Optional(
            "joboutputdir", default=lambda: os.path.join(APP_DIR, "joboutput")
        ): os.path.exists,
        Optional("history_length", default=1000): int,
        Optional("slurm_driver"): slurm_schema,
        Optional("htcondor_driver"): htcondor_schema,
        Optional("notify", default=[]): [{"name": str, Optional(object): object}],
        Optional(object): object,
    }
)


class Notifier:
    """
    Class that manages a notification provider. This essentially wraps :class:`notifiers.core.Provider` and enables
    a bit more automatic filling of fields depending on the accepted provider schema.
    """

    name: typing.Optional[str] = None
    args: typing.Optional[typing.Dict[str, typing.Any]] = None
    _notifier: notifiers.core.Provider

    def __init__(self, name: str, **kwargs: typing.Any):
        """
        Initialize method for the generic notifier
        :param name: The name of this notifier, is used to instantiate the underlying provider
        :param kwargs: Any additional arguments to be passed to the provider at construction
        """
        from .logger import logger

        self.logger = logger
        self.name = name
        self._notifier = notifiers.get_notifier(self.name)
        self._kwargs = kwargs

    def notify(
        self, message: str, title: typing.Optional[str] = None, **kwargs: typing.Any,
    ) -> notifiers.core.Response:
        """
        Send a notification through this notificer instance

        .. note::
           If you specify a title it will be set as the title or subject field if the provider
           supports it, otherwise it will be prepended to the message

        :param message: The message to send
        :param title: A title to send, optional
        :param kwargs: Any additional keyword arguments to pass to the provider's notify call
        :return: a :class:`notifiers.core.Response` instance
        """
        self.logger.debug("Sending notification '%s' via %s", message, self.name)

        kwargs = kwargs.copy()
        kwargs["message"] = f"{socket.gethostname()}: {message}"
        if title is not None:
            if "title" in self._notifier.schema["properties"]:
                kwargs["title"] = title
            elif "subject" in self._notifier.schema["properties"]:
                kwargs["subject"] = title
            else:
                kwargs["message"] = title + ":\n" + kwargs["message"]

        return self._notifier.notify(**self._kwargs, **kwargs)


class NotificationManager:
    """
    Class to group and handle multiple notifiers/providers. Will always send to all of them.
    """

    notifiers: typing.List[Notifier]

    def __init__(self, config: "Config"):
        """
        Initializer for the notification manager
        :param config: Config object that is used to configure the :class:`Notifier` instances.
        """
        from .logger import logger

        self.logger = logger

        self.notifiers = []
        if "notify" in config.data and config.data["notify"] is not None:
            for item in config.data["notify"]:
                name = item["name"]
                args = dict(item)
                del args["name"]
                self.notifiers.append(Notifier(name=name, **args))

    def notify(
        self,
        message: str,
        title: typing.Optional[str] = None,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.List[notifiers.core.Response]:  # type: ignore
        """
        Sends a notification using all the configured notifiers/providers.

        .. note::
           If you specify a title it will be set as the title or subject field if the provider
           supports it, otherwise it will be prepended to the message

        :param message: The message to send
        :param title: The title for the notification
        :param args: Any additional positional arguments to pass to the providers
        :param kwargs: Any additional keyword arguments to pass to the providers
        :return: List of :class:`notifiers.core.Response` instances
        """
        self.logger.debug("%d notifiers configured", len(self.notifiers))

        responses: typing.List[notifiers.core.Response] = []
        for notifier in self.notifiers:
            responses.append(notifier.notify(message, title, *args, **kwargs))
        return responses

    @property
    def enabled(self) -> bool:
        return len(self.notifiers) > 0


class Config:
    """
    Class to handle loading the config data from disk.
    """

    def __init__(
        self, data: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> None:
        """
        Initalize method for the config. Will load the config file from the app directory (OS dependant)

        Parameters
        ----------
        data
            Dictionary with pre-loaded data. Will be used as is if provided (optional)
        """

        if data is not None:
            self.data = data
        else:
            with open(CONFIG_FILE) as f:
                self.data = yaml.safe_load(f)

        self.data = config_schema.validate(self.data)

        self.notifications = NotificationManager(self)

    def __getattr__(self, key: str) -> typing.Any:
        if key not in self.data:
            raise AttributeError()
        return self.data[key]



import os
import re
import tempfile
import time
from concurrent.futures import Executor, Future
from datetime import timedelta

import pytest
from unittest import mock
from unittest.mock import Mock, ANY, MagicMock
import peewee as pw
from click import UsageError
from conftest import skip_lxplus

from kong.model.folder import Folder
from kong.model.job import Job

from kong.repl import Repl, complete_path
import kong

import logging

from kong.state import DoesNotExist

kong.logger.logger.setLevel(logging.DEBUG)


@pytest.fixture
def repl(state):
    r = Repl(state)
    r._raise = True
    return r


def test_ls(tree, state, repl, capsys, sample_jobs, monkeypatch):
    repl.do_ls(".")
    out, err = capsys.readouterr()
    assert all(f.name in out for f in state.cwd.children)
    assert all(f"{j.job_id}" in out for j in state.cwd.jobs)

    state.mkdir("f4")
    repl.onecmd("ls f4")

    j1 = Job.get(1)
    j1.batch_job_id = None
    j1.save()
    state.mv(j1, "f1")

    repl.onecmd("ls")
    out, err = capsys.readouterr()
    assert all(f.name in out for f in state.cwd.children)
    assert all(f"{j}" in out for j in [2, 3])

    repl.onecmd("ls --recursive")
    out, err = capsys.readouterr()
    # assert all(f.name in out for f in state.cwd.children)
    all_jobs = [j.job_id for j in Job.select()]
    assert all(f"{j}" in out for j in all_jobs)

    repl.do_ls("/nope")
    out, err = capsys.readouterr()
    assert "not exist" in out

    state.cwd = Folder.find_by_path("/f2", state.cwd)
    repl.do_ls(".")
    out, err = capsys.readouterr()
    assert all(f.name in out for f in state.cwd.children)
    assert all(f"{j.job_id}" in out for j in state.cwd.jobs)

    with pytest.raises(UsageError):
        repl.onecmd("ls --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out

    with monkeypatch.context() as m:

        job = state.create_job(command="sleep 1")
        job.batch_job_id = None

        m.setattr(state, "refresh_jobs", Mock(return_value=[]))
        m.setattr(state, "ls", Mock(return_value=([], [job])))

        repl.onecmd("ls -R /")

        state.ls.assert_called_once()
        assert state.refresh_jobs.call_count == 1

def test_ls_status(state, repl, capsys, monkeypatch):
    monkeypatch.setattr("kong.repl.Spinner", MagicMock())

    j1 = state.create_job(command="sleep 1")
    j2 = state.create_job(command="sleep 1")
    j3 = state.create_job(command="sleep 1")

    j2.status = Job.Status.FAILED
    j2.save()

    repl.onecmd("ls")
    out, err = capsys.readouterr()
    assert len(out.strip().split("\n")) == 5

    repl.onecmd("ls -S CREATED")
    out, err = capsys.readouterr()
    assert len(out.strip().split("\n")) == 4

    repl.onecmd("ls -S FAILED")
    out, err = capsys.readouterr()
    assert len(out.strip().split("\n")) == 3



def test_ls_sizes(db, tree, state, repl, capsys, sample_jobs, monkeypatch):
    class DirectExecutor(Executor):
        def __init__(self, *args, **kwargs):
            pass

        def submit(self, fn, *args, **kwargs):
            future = Future()
            try:
                future.set_result(fn(*args, **kwargs))
            except Exception as e:
                future.set_exception(e)
            return future

        def __enter__(self):
            return self

    monkeypatch.setattr(
        "kong.repl.ThreadPoolExecutor", DirectExecutor
    )  # disable threads
    monkeypatch.setattr("kong.repl.Job.size", Mock(return_value=42))
    monkeypatch.setattr("kong.repl.Spinner", MagicMock())
    repl.onecmd("ls -s .")
    out, err = capsys.readouterr()
    lines = out.strip().split("\n")
    exp = """
name output size              UNKNOWN CREATED SUBMITTED RUNNING FAILED COMPLETED
---- ------------------------ ------- ------- --------- ------- ------ ---------
f1   168 bytes                      0       4         0       0      0         0
f2   252 bytes                      0       6         0       0      0         0
f3   0 bytes                        0       0         0       0      0         0
"""[
        1:-1
    ]
    assert "\n".join(lines[:6]).strip() == exp


@skip_lxplus
def test_ls_refresh(repl, state, capsys, sample_jobs, monkeypatch):
    repl.do_ls(".")
    out, err = capsys.readouterr()
    assert "CREATED" in out

    for job in sample_jobs:
        job.submit()

    time.sleep(0.2)

    # without refresh
    repl.do_ls(".")
    out, err = capsys.readouterr()
    lines = out.split("\n")[:-1]
    assert all("SUBMITTED" in l for l in lines[-3:])
    assert all("COMPLETED" not in l for l in lines[-3:])

    # with refresh
    repl.do_ls(". --refresh")
    out, err = capsys.readouterr()
    lines = out.split("\n")[:-1]
    assert all("SUBMITTED" not in l for l in lines[-3:])
    assert all("COMPLETED" in l for l in lines[-3:])

    with monkeypatch.context() as m:
        mock = Mock(return_value=[])
        m.setattr(state, "refresh_jobs", mock)
        repl.onecmd("ls -R /")
        assert mock.call_count == 2 # called once for folders, once for jobs


def test_complete_path(state, tree, repl):
    root = Folder.get_root()
    alts = complete_path(root, "f")
    assert alts == ["f1/", "f2/", "f3/"]

    alts = complete_path(root, "f1")
    assert alts == ["f1/"]

    alts = complete_path(root, "f2/")
    assert alts == ["alpha/", "beta/", "gamma/"]

    state.cwd = Folder.find_by_path("/f2", state.cwd)

    alts = complete_path(root.subfolder("f2"), "a")
    assert alts == ["alpha/"]

def test_completed_default(repl):
    root = Folder.get_root()
    root.add_folder("alpha")
    root.add_folder("beta_delta")
    root.add_folder("beta_gamma")

    assert repl.completedefault("al", "ls al", 3, 5) == ["alpha/"]
    assert repl.completedefault("be", "ls alpha be", 9, 11) == ["beta_delta/", "beta_gamma/"]
    assert repl.completedefault("alp", "ls alp beta_delta", 3, 6) == ["alpha/"]
    assert repl.completedefault("alp", "ls alp nope", 3, 6) == ["alpha/"]
    assert repl.completedefault("be", "ls be", 3, 5) == ["beta_delta/", "beta_gamma/"]


def test_mkdir(state, repl, db, capsys, monkeypatch):
    root = Folder.get_root()
    sub = root.add_folder("sub")

    for cwd in [root, sub]:
        state.cwd = cwd

        assert cwd.subfolder("alpha") is None
        repl.do_mkdir("alpha")
        out, err = capsys.readouterr()
        alpha = cwd.subfolder("alpha")
        assert alpha is not None

        # one down
        assert alpha.subfolder("beta") is None
        repl.do_mkdir("alpha/beta")
        out, err = capsys.readouterr()
        beta = alpha.subfolder("beta")
        assert beta is not None

        # cannot create outside of root
        repl.do_mkdir("../nope")
        if cwd == root:
            out, err = capsys.readouterr()
            assert "annot create" in out and "../nope" in out
        else:
            assert root.subfolder("nope") is not None

        # cannot create again
        repl.do_mkdir("alpha")
        out, err = capsys.readouterr()
        assert "alpha" in out

        # cannot create in nonexistant
        repl.do_mkdir("omega/game")
        out, err = capsys.readouterr()
        assert "omega/game" in out and "annot create" in out

        state.cwd = beta
        assert cwd.subfolder("gamma") is None
        repl.do_mkdir("../../gamma")
        out, err = capsys.readouterr()
        gamma = cwd.subfolder("gamma")
        assert gamma is not None

        # force integrity error
        with monkeypatch.context() as m:
            m.setattr(state, "mkdir", Mock(side_effect=pw.IntegrityError))
            repl.onecmd("mkdir hurz")
            out, err = capsys.readouterr()
            assert "already exists" in out

    assert Repl.do_mkdir.__doc__ is not None


def test_mkdir_create_parents(state, repl, capsys):
    root = Folder.get_root()

    repl.onecmd("mkdir /a1/b2/c3/d4")
    out, err = capsys.readouterr()
    assert "Cannot create folder" in out
    assert Folder.find_by_path("/a1/b2/c3/d4", state.cwd) is None

    repl.onecmd("mkdir -p /a1/b2/c3/d4")
    out, err = capsys.readouterr()
    assert Folder.find_by_path("/a1", state.cwd) is not None
    assert Folder.find_by_path("/a1/b2", state.cwd) is not None
    assert Folder.find_by_path("/a1/b2/c3", state.cwd) is not None
    assert Folder.find_by_path("/a1/b2/c3/d4", state.cwd) is not None



def test_cd(state, repl, db, capsys):
    root = Folder.get_root()
    assert state.cwd == root

    repl.do_cd("nope")
    out, err = capsys.readouterr()
    assert "not exist" in out, "nope" in out
    assert state.cwd == root

    nope = root.add_folder("nope")
    repl.do_cd("nope")
    out, err = capsys.readouterr()
    assert state.cwd == nope

    repl.do_cd("")
    out, err = capsys.readouterr()
    assert state.cwd == root

    repl.do_cd("..")
    out, err = capsys.readouterr()
    assert "not exist" in out
    assert state.cwd == root

    repl.do_cd("../nope")
    out, err = capsys.readouterr()
    assert "not exist" in out
    assert state.cwd == root

    more = root.add_folder("more")
    another = nope.add_folder("another")

    repl.do_cd("/nope")
    out, err = capsys.readouterr()
    assert state.cwd == nope

    repl.do_cd("/nope/another")
    out, err = capsys.readouterr()
    assert state.cwd == another

    repl.do_cd("/../")
    out, err = capsys.readouterr()
    assert state.cwd == another

    repl.do_cd("..")
    out, err = capsys.readouterr()
    assert state.cwd == nope

    repl.do_cd("/more")
    out, err = capsys.readouterr()
    assert state.cwd == more


def test_mv_folder(state, repl, capsys):
    root = Folder.get_root()

    repl.onecmd("mv --help")
    out, err = capsys.readouterr()

    f1, f2, f3, f4, f5 = [root.add_folder(n) for n in ("f1", "f2", "f3", "f4", "f5")]

    assert len(root.children) == 5
    assert len(f2.children) == 0

    # actual move
    repl.onecmd("mv f1 f2")
    assert len(root.children) == 4
    f2.reload()
    assert len(f2.children) == 1 and f2.children[0] == f1
    f1.reload()
    assert f1.parent == f2
    out, err = capsys.readouterr()

    # rename f3 -> f3x
    repl.onecmd("mv f3 f3x")
    out, err = capsys.readouterr()
    f3.reload()
    assert len(root.children) == 4
    assert f3.name == "f3x"

    # another move
    repl.onecmd("mv f3x f4")
    assert len(f4.children) == 1 and f4.children[0] == f3
    f3.reload()
    assert f3.parent == f4
    assert f3.name == "f3x"
    out, err = capsys.readouterr()

    # move rename at the same time
    repl.onecmd("cd f2")
    repl.onecmd("mv ../f5 ../f4/f5x")
    out, err = capsys.readouterr()
    f5.reload()
    assert len(f4.children) == 2
    assert f5.name == "f5x"
    assert f5.parent == f4

    # try move to nonexistant
    repl.onecmd("cd /")
    with pytest.raises(ValueError):
        repl.onecmd("mv f2/f1 /nope/blub")
    out, err = capsys.readouterr()
    assert "/nope" in out and "not exist" in out

    # try to move nonexistant
    with pytest.raises(DoesNotExist):
        repl.onecmd("mv ../nope f1")
    out, err = capsys.readouterr()
    assert "../nope" in out and "No such" in out


def test_mv_job(state, repl, capsys):
    root = Folder.get_root()

    f1, f2 = [root.add_folder(n) for n in ("f1", "f2")]

    assert len(root.children) == 2

    j1, j2, j3, j4, j5 = [state.create_job(command="sleep 1") for _ in range(5)]
    assert len(root.jobs) == 5

    repl.onecmd(f"mv {j1.job_id} f1")
    j1.reload()
    assert j1.folder == f1
    assert len(f1.jobs) == 1
    assert len(root.jobs) == 4
    out, err = capsys.readouterr()

    repl.onecmd(f"mv {j2.job_id} f2")
    j2.reload()
    assert j2.folder == f2
    assert len(f2.jobs) == 1
    assert len(root.jobs) == 3
    out, err = capsys.readouterr()

    state.cd(f2)
    repl.onecmd(f"mv {j3.job_id} .")
    j3.reload()
    assert j3.folder == f2
    out, err = capsys.readouterr()

    repl.onecmd(f"mv {j2.job_id} ..")
    j2.reload()
    assert j2.folder == root
    out, err = capsys.readouterr()

    repl.onecmd(f"mv ../{j4.job_id} ../f1")
    j4.reload()
    assert j4.folder == f1
    out, err = capsys.readouterr()

    state.cd(root)

    # renaming does not work
    with pytest.raises(ValueError):
        repl.onecmd(f"mv {j5.job_id} 42")
    out, err = capsys.readouterr()
    assert "42" in out and "not exist" in out


def test_mv_bulk_job(state, repl):
    root = Folder.get_root()

    f1, f2 = [root.add_folder(n) for n in ("f1", "f2")]
    assert len(root.children) == 2

    j1, j2, j3, j4, j5 = [state.create_job(command="sleep 1") for _ in range(5)]
    assert len(root.jobs) == 5

    repl.onecmd("mv * f1")
    assert len(root.jobs) == 0 and len(f1.jobs) == 5
    assert len(root.children) == 1  # Didn't move the folders
    for j in (j1, j2, j3, j4, j5):
        j.reload()
        assert j.folder == f1

    state.cwd = f2
    repl.onecmd("mv ../f1/* .")
    assert len(f1.jobs) == 0 and len(f2.jobs) == 5
    for j in (j1, j2, j3, j4, j5):
        j.reload()
        assert j.folder == f2


def test_mv_bulk_folder(state, repl):
    root = Folder.get_root()

    r1, r2 = [root.add_folder(n) for n in ("r1", "r2")]

    folders = [r1.add_folder(f"f{n}") for n in range(5)]
    assert len(r1.children) == len(folders)
    assert len(r2.children) == 0

    repl.onecmd("mv r1/* r2")
    assert len(r1.children) == 0
    assert len(r2.children) == len(folders)
    for f in folders:
        f.reload()
        assert f.parent == r2

    state.cwd = r1
    repl.onecmd("mv ../r2/* .")
    assert len(r1.children) == len(folders)
    assert len(r2.children) == 0
    for f in folders:
        f.reload()
        assert f.parent == r1


def test_mv_error(state, repl, capsys, monkeypatch):
    with pytest.raises(UsageError):
        repl.onecmd("mv --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


def test_rm(state, repl, db, capsys, monkeypatch):
    root = Folder.get_root()

    repl.do_rm("../nope")
    out, err = capsys.readouterr()
    assert "not exist" in out

    repl.do_rm("/")
    out, err = capsys.readouterr()
    assert "annot delete" in out, "root" in out

    root.add_folder("alpha")
    with monkeypatch.context() as m:
        confirm = Mock(return_value=False)
        m.setattr("click.confirm", confirm)
        repl.onecmd("rm -r alpha")
        confirm.assert_called_once()

    assert root.subfolder("alpha") is not None
    with monkeypatch.context() as m:
        confirm = Mock(return_value=True)
        m.setattr("click.confirm", confirm)
        repl.onecmd("rm -r alpha")
        confirm.assert_called_once()
    assert root.subfolder("alpha") is None
    out, err = capsys.readouterr()
    assert len(out) > 0


def test_rm_job(state, repl, db, capsys, monkeypatch):
    root = Folder.get_root()
    j1 = state.default_driver.create_job(command="sleep 1", folder=root)
    assert len(root.jobs) == 1 and root.jobs[0] == j1
    assert Job.get_or_none(job_id=j1.job_id) is not None

    with monkeypatch.context() as m:
        confirm = Mock(return_value=True)
        m.setattr("click.confirm", confirm)
        repl.do_rm(str(j1.job_id))
        confirm.assert_called_once()

    out, err = capsys.readouterr()
    assert len(root.jobs) == 0
    assert Job.get_or_none(job_id=j1.job_id) is None

    # works in other cwd too
    alpha = root.add_folder("alpha")
    j2 = state.default_driver.create_job(command="sleep 1", folder=alpha)
    assert j1.job_id != j2.job_id
    assert Job.get_or_none(job_id=j2.job_id) is not None
    assert len(alpha.jobs) == 1 and alpha.jobs[0] == j2
    assert state.cwd == root
    with monkeypatch.context() as m:
        confirm = Mock(return_value=True)
        m.setattr("click.confirm", confirm)
        repl.do_rm(str(j2.job_id))
        confirm.assert_called_once()
    out, err = capsys.readouterr()
    assert Job.get_or_none(job_id=j2.job_id) is None
    assert len(alpha.jobs) == 0


def test_cwd(state, repl, tree, capsys):
    root = tree
    for cmd in ("cwd", "pwd"):
        state.cwd = root
        repl.onecmd(cmd)
        out, err = capsys.readouterr()
        assert out.strip() == "/"

        state.cwd = root / "f1"
        repl.onecmd(cmd)
        out, err = capsys.readouterr()
        assert out.strip() == "/f1"

        state.cwd = root / "f2" / "gamma"
        repl.onecmd(cmd)
        out, err = capsys.readouterr()
        assert out.strip() == "/f2/gamma"


def test_wait(repl, state, monkeypatch):
    wait = Mock(return_value=iter([]))
    monkeypatch.setattr(state, "wait", wait)

    repl.onecmd("wait * --no-notify --recursive --poll-interval 50")
    wait.assert_called_once_with(
        ("*",), notify=False, recursive=True, poll_interval=50, update_interval=None
    )

    wait.reset_mock()

    repl.onecmd("wait * --notify --recursive --poll-interval 50 --notify-interval 30m")
    wait.assert_called_once_with(
        ("*",),
        notify=True,
        recursive=True,
        poll_interval=50,
        update_interval=timedelta(minutes=30),
    )



def test_exit(repl):
    assert repl.do_exit("") == True
    assert repl.do_EOF("") == True


def test_preloop(repl, monkeypatch):
    m = Mock()
    monkeypatch.setattr("readline.read_history_file", m)
    monkeypatch.setattr("os.path.exists", Mock(return_value=True))
    repl.preloop()
    m.assert_called_once()
    monkeypatch.setattr("os.path.exists", Mock(return_value=False))
    repl.preloop()


def test_postloop(state, repl, monkeypatch):
    repl.postloop()


def test_precmd(repl):
    assert repl.precmd("whatever") == "whatever"
    assert repl.precmd("") == ""


def test_onecmd(repl, monkeypatch, capsys):
    set_length = Mock()
    write = Mock()
    monkeypatch.setattr("readline.set_history_length", set_length)
    monkeypatch.setattr("readline.write_history_file", write)
    m = Mock(return_value="ok")
    monkeypatch.setattr("cmd.Cmd.onecmd", m)
    assert repl.onecmd("whatever") == "ok"
    set_length.assert_called_once_with(repl.state.config.history_length)
    write.assert_called_once()
    m.assert_called_once()

    m = Mock(side_effect=TypeError("MESSAGE"))
    monkeypatch.setattr("cmd.Cmd.onecmd", m)
    with pytest.raises(TypeError):
        repl.onecmd("whatever")
    out, err = capsys.readouterr()
    assert "MESSAGE" in out
    m = Mock(side_effect=RuntimeError())
    monkeypatch.setattr("cmd.Cmd.onecmd", m)
    with monkeypatch.context() as m:
        m.setattr(repl, "_raise", False) # disable debug mode for this check
        repl.onecmd("whatever")  # swallows other exceptions


def test_cmdloop(repl, monkeypatch, capsys):
    m = Mock(return_value="ok")
    monkeypatch.setattr("cmd.Cmd.cmdloop", m)
    repl.cmdloop()
    m.assert_called_once()

    m = Mock(side_effect=[KeyboardInterrupt(), "ok"])
    monkeypatch.setattr("cmd.Cmd.cmdloop", m)
    repl.cmdloop()
    out, err = capsys.readouterr()
    m.assert_called()
    assert m.call_count == 2
    assert "^C" in out


def test_emptyline(repl):
    repl.emptyline()


def test_create_job(repl, state, tree, capsys):
    root = tree

    repl.do_create_job("")
    out, err = capsys.readouterr()
    assert "provide a command" in out

    cmd = "sleep 1"
    repl.do_create_job(f"{cmd}")

    assert len(root.jobs) == 1
    j1 = root.jobs[-1]
    assert j1.command == cmd
    assert j1.status == Job.Status.CREATED
    out, err = capsys.readouterr()
    assert "reated" in out
    assert str(j1.job_id) in out
    assert j1.batch_job_id in out

    repl.do_create_job("--help")
    out, err = capsys.readouterr()
    assert "Usage" in out

    with pytest.raises(UsageError):
        repl.onecmd("create_job --nope 5 sleep 2")  # wrong option --core
    out, err = capsys.readouterr()
    assert "no such" in out

    repl.onecmd("create_job -- exe --and --some arguments --and options")
    out, err = capsys.readouterr()

    j4 = root.jobs[-1]
    assert j4.command == "exe --and --some arguments --and options"


def test_create_job_extra_arguments(repl, state, tree, monkeypatch):
    root = tree

    cmd = "sleep 1"
    cores = 16

    monkeypatch.setattr(state, "create_job", Mock())
    repl.onecmd(f"create_job -a cores={cores} '{cmd}'")
    assert state.create_job.call_count == 1
    kwargs = state.create_job.call_args[1]
    assert kwargs["cores"] == cores

    state.create_job.reset_mock()
    repl.onecmd(f"create_job -a ARGA=hurz --argument ARGB=blurz -a ARGC=42 '{cmd}'")
    assert state.create_job.call_count == 1
    kwargs = state.create_job.call_args[1]
    print(kwargs)
    assert kwargs["ARGA"] == "hurz"
    assert kwargs["ARGB"] == "blurz"
    assert kwargs["ARGC"] == 42


@skip_lxplus
def test_submit_job(repl, state, capsys, monkeypatch):
    root = Folder.get_root()
    value = "VALUE VALUE VALUE"
    cmd = f"sleep 0.3; echo '{value}'"

    monkeypatch.setattr("click.confirm", Mock(return_value=True))

    repl.do_create_job(cmd)
    j1 = root.jobs[-1]
    j1.ensure_driver_instance(state.config)
    assert j1.status == Job.Status.CREATED

    repl.do_submit_job(f"{j1.job_id}")
    j1.reload()
    assert j1.status == Job.Status.SUBMITTED
    time.sleep(0.1)
    assert j1.get_status() == Job.Status.RUNNING
    time.sleep(0.3)
    assert j1.get_status() == Job.Status.COMPLETED

    out, err = capsys.readouterr()

    with pytest.raises(UsageError):
        repl.onecmd("submit_job --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


def test_kill_job(repl, state, capsys, monkeypatch):
    root = Folder.get_root()
    repl.do_create_job("sleep 1")
    j1 = root.jobs[-1]
    j1.ensure_driver_instance(state.config)

    assert j1.status == Job.Status.CREATED
    j1.submit()
    assert j1.status == Job.Status.SUBMITTED
    time.sleep(0.1)
    assert j1.get_status() == Job.Status.RUNNING

    monkeypatch.setattr("click.confirm", Mock(return_value=True))
    repl.do_kill_job(str(j1.job_id))
    out, err = capsys.readouterr()
    assert j1.get_status() == Job.Status.FAILED

    with pytest.raises(UsageError):
        repl.onecmd("kill_job --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


def test_resubmit_job(repl, state, capsys, monkeypatch):
    root = Folder.get_root()
    repl.do_create_job("sleep 1")
    j1 = root.jobs[-1]
    j1.ensure_driver_instance(state.config)

    assert j1.status == Job.Status.CREATED
    j1.submit()
    assert j1.status == Job.Status.SUBMITTED
    time.sleep(0.1)
    assert j1.get_status() == Job.Status.RUNNING

    monkeypatch.setattr("click.confirm", Mock(return_value=True))
    repl.do_kill_job(str(j1.job_id))

    assert j1.get_status() == Job.Status.FAILED

    repl.do_resubmit_job(str(j1.job_id))
    out, err = capsys.readouterr()
    j1.reload()
    assert j1.status == Job.Status.SUBMITTED

    with pytest.raises(UsageError):
        repl.onecmd("resubmit_job --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


@skip_lxplus
def test_update(repl, state, capsys, monkeypatch):
    root = Folder.get_root()

    repl.onecmd("update")
    out, err = capsys.readouterr()

    repl.onecmd("update --help")
    out, err = capsys.readouterr()
    assert "Usage" in out

    with pytest.raises(DoesNotExist):
        repl.onecmd("update 42")
    out, err = capsys.readouterr()
    assert "not find" in out

    j1 = state.create_job(command="sleep 0.2")

    def update():
        repl.onecmd(f"update {j1.job_id}")
        out, err = capsys.readouterr()

    update()

    j1.reload()
    assert j1.status == Job.Status.CREATED

    j1.submit()
    assert j1.status == Job.Status.SUBMITTED

    time.sleep(0.1)

    update()
    j1.reload()
    assert j1.status == Job.Status.RUNNING

    time.sleep(0.2)

    update()
    j1.reload()
    assert j1.status == Job.Status.COMPLETED

    with pytest.raises(UsageError):
        repl.onecmd("update --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out

    j2 = state.create_job(command="sleep 0.2")

    repl.onecmd("update -r .")
    out, err = capsys.readouterr()
    lines = out.strip("").split("\n")
    # assert lines[2] == "bla"
    assert re.match(r"\s*0U\s*1C\s*0S\s*0R\s*0F\s*1C", lines[2]) is not None


def test_info(state, repl, capsys, monkeypatch):
    repl.onecmd("info") # missing arg
    out, err = capsys.readouterr()
    assert "usage" in out


    job = state.create_job(command="sleep 1")

    repl.onecmd("info 1")
    out, err = capsys.readouterr()
    assert str(job.job_id) in out
    assert job.batch_job_id in out
    for k, v in job.data.items():
        assert k in out
        assert v in out

    with monkeypatch.context() as m:
        refresh = Mock(return_value=[])
        m.setattr(state, "refresh_jobs", refresh)
        repl.onecmd("info 1 -R")
        out, err = capsys.readouterr()
        refresh.assert_called_once()

    command_string = "X" * 600

    long_job = state.create_job(command=command_string)

    repl.onecmd("info 2")
    out, err = capsys.readouterr()
    assert not command_string in out

    repl.onecmd("info 2 --full")
    out, err = capsys.readouterr()
    assert command_string in out


def test_tail(state, repl, capsys, monkeypatch):
    job = state.create_job(command="sleep 1")

    with monkeypatch.context() as m:
        tail = Mock()
        m.setattr("sh.tail", tail)
        m.setattr("time.sleep", Mock())
        m.setattr("os.path.exists", Mock(side_effect=[False, False, True]))
        spinner = MagicMock()
        m.setattr("kong.repl.Spinner", spinner)
        repl.onecmd(f"tail {job.job_id}")
        out, err = capsys.readouterr()
        spinner.assert_called_once()
        assert tail.call_count == 1

    with monkeypatch.context() as m:
        tail = Mock()
        m.setattr("sh.tail", tail)
        m.setattr("os.path.exists", Mock(side_effect=[True]))
        spinner = MagicMock()
        m.setattr("kong.repl.Spinner", spinner)
        repl.onecmd(f"tail {job.job_id}")
        out, err = capsys.readouterr()
        assert spinner.call_count == 0
        assert tail.call_count == 1

    with pytest.raises(UsageError):
        repl.onecmd(f"tail --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


def test_less(state, repl, capsys, monkeypatch):
    job = state.create_job(command="sleep 1")

    content = "SOMECONTENT: BLABLBALBALBALBLA\nNEWLINE"
    with tempfile.NamedTemporaryFile("wt") as f:
        f.write(content)
        f.flush()
        job.data["stdout"] = f.name
        job.save()

        lines = []

        def agg(it):
            nonlocal lines
            lines = list(it)

        with monkeypatch.context() as m:
            pager = Mock(side_effect=agg)
            m.setattr("click.echo_via_pager", pager)
            repl.onecmd(f"less {job.job_id}")
            out, err = capsys.readouterr()
            pager.assert_called_once()
            assert "".join(lines) == content

    with pytest.raises(UsageError):
        repl.onecmd(f"less --nope")
    out, err = capsys.readouterr()
    assert "no such option" in out


def test_shell(repl, monkeypatch):
    string = "HALLO HALLO 123"
    cmd = f"echo '{string}'"

    run = Mock()
    monkeypatch.setattr("subprocess.run", run)
    repl.onecmd(f"!{cmd}")
    run.assert_called_once_with(cmd, shell=True, env=ANY)



import os

import pytest
import peewee as pw

from kong.db import database
from kong.model.folder import Folder

import logging

from kong.model.job import Job

logging.getLogger("kong").setLevel(logging.DEBUG)


def test_create(db):
    root = Folder.get_root()
    f1 = Folder.create(name="hallo", parent=root)

    assert f1.parent is root
    assert len(root.children) == 1

    f2 = Folder.create(name="number2", parent=f1)
    assert len(root.children) == 1
    assert f2.name == "number2"
    assert f2.parent == f1
    assert f2.parent_id == f1.folder_id

    with pytest.raises(AssertionError):
        Folder.create(name=".")
    with pytest.raises(AssertionError):
        Folder.create(name="..")
    with pytest.raises(AssertionError):
        Folder.create(name="/")

    f3 = root.add_folder("f3")
    assert f3.name == "f3"
    assert f3.parent == root
    assert len(root.children) == 2


def test_cannot_be_own_parent(db):
    root = Folder.get_root()
    f1 = root.add_folder("f1")
    assert f1.parent == root

    with pytest.raises(pw.IntegrityError):
        f1.parent = f1
        f1.save()
    f1.reload()
    assert f1.parent == root


def test_create_name_unique(db):
    root = Folder.get_root()
    f1 = root.add_folder("f1")

    with pytest.raises(pw.IntegrityError):
        Folder.create(name="f1", parent=root)
    with pytest.raises(pw.IntegrityError):
        root.add_folder("f1")

    # another one is fine
    assert root.add_folder("f2").name == "f2"
    # one further down is also fine
    assert f1.add_folder("f1").name == "f1"


def test_create_name_validate(db):
    root = Folder.get_root()
    with pytest.raises(AssertionError):
        Folder.create(name="123", parent=root)


def test_get_subfolder(db):
    root = Folder.get_root()

    f1 = root.add_folder("f1")
    f2 = root.add_folder("f2")
    f2_f1 = f2.add_folder("f1")

    assert root.subfolder("f1") == f1
    assert root.subfolder("f2") == f2
    assert f2.subfolder("f1") == f2_f1


def test_get_root_creates(db):
    assert Folder.get_or_none(name="root", parent=None) is None

    root = Folder.get_root()
    assert root.name == "root"
    assert root.parent is None

    assert Folder.get_or_none(name="root", parent=None) is not None


def test_only_one_root(db):
    root = Folder.get_root()
    # try to create another one

    with pytest.raises(AssertionError):
        Folder.create(name="something else")
    with pytest.raises(AssertionError):
        Folder.create(name="something else", parent=None)

    # make sure it wasnt created
    assert Folder.select().where(Folder.parent == None).count() == 1


def test_path(db):
    root = Folder.get_root()
    assert root.path == "/"

    f1 = Folder.create(name="f1", parent=root)
    assert f1.path == "/f1"
    f2 = Folder.create(name="f2", parent=f1)
    assert f2.path == "/f1/f2"
    f3 = Folder.create(name="f3", parent=f2)
    assert f3.path == "/f1/f2/f3"
    f4 = Folder.create(name="f4", parent=f3)
    assert f4.path == "/f1/f2/f3/f4"

    # move stuff around
    f4.parent = f2
    f4.save()
    assert f4.path == "/f1/f2/f4"


def test_find_by_path(db):
    root = Folder.get_root()

    f1 = root.add_folder("f1")
    f2 = f1.add_folder("f2")
    f3 = f2.add_folder("f3")
    f4 = f2.add_folder("f4")


    # Check it works with implicit CWD
    assert Folder.find_by_path("/") == root
    assert Folder.find_by_path("../blubb") == None
    assert Folder.find_by_path("/f1") == f1
    assert Folder.find_by_path("/f1/f2") == f2
    assert Folder.find_by_path("/f1/f2/f3") == f3
    assert Folder.find_by_path("/f1/f2/f4") == f4

    for f in [root, f1, f2, f3, f4]:
        # absolute paths work regardless of cwd
        assert Folder.find_by_path("/", f) == root
        assert Folder.find_by_path("/f1", f) == f1
        assert Folder.find_by_path("/f1/f2", f) == f2
        assert Folder.find_by_path("/f1/f2/f3", f) == f3
        assert Folder.find_by_path("/f1/f2/f4", f) == f4
        # self referential paths work everywhere
        assert Folder.find_by_path(".", f) == f
        assert Folder.find_by_path(f.path, f) == f
        assert Folder.find_by_path(f.path + "/", f) == f

        # for root, it's None
        assert Folder.find_by_path("..", f) == f.parent

        assert Folder.find_by_path("nope", f) is None
        assert Folder.find_by_path("../nope", f) is None

    assert Folder.find_by_path("f1", root) == f1
    assert Folder.find_by_path("f1") == f1
    assert Folder.find_by_path("../", f1) == root
    assert Folder.find_by_path("../f1", f1) == f1
    assert Folder.find_by_path("../f1/f2", f1) == f2
    assert Folder.find_by_path("f2", f1) == f2
    assert Folder.find_by_path("f3", f2) == f3
    assert Folder.find_by_path("f4", f2) == f4
    assert Folder.find_by_path("../f4", f3) == f4
    assert Folder.find_by_path("../f3", f4) == f3


def test_fancy_operator(db):
    root = Folder.get_root()

    f1 = root.add_folder("f1")
    f2 = f1.add_folder("f2")
    f3 = f2.add_folder("f3")
    f4 = f2.add_folder("f4")

    assert root / "f1" == f1
    assert root / "f1" / "f2" == f2
    assert root / "f1" / "f2" / "f3" == f3
    assert root / "f1" / "f2" / "f4" == f4

def test_folders_recursive(db, state):
    root = Folder.get_root()
    f1 = root.add_folder("f1")
    f2 = f1.add_folder("f2")
    f3 = f1.add_folder("f3")
    f4 = root.add_folder("f4")

    folders = root.folders_recursive()

    assert set(folders) == set([f1, f2, f3, f4])


def test_jobs_recursive(db, state):
    root = Folder.get_root()
    f1 = root.add_folder("f1")
    f2 = f1.add_folder("f2")

    with state.pushd(f1):
        j1 = state.create_job(command="sleep 1")
    with state.pushd(f2):
        j2 = state.create_job(command="sleep 1")

    jobs = root.jobs_recursive()
    assert len(jobs) == 2
    assert all(a == b for a, b in zip(jobs, [j1, j2]))

def test_job_stats(db, state):

    root = Folder.get_root()

    f1 = root.add_folder("f1")
    f2 = f1.add_folder("f2")

    state.cd(f1)
    j1 = state.create_job(command="sleep 1")
    j2 = state.create_job(command="sleep 1")
    j3 = state.create_job(command="sleep 1")

    state.cd(f2)
    j4 = state.create_job(command="sleep 1")
    j5 = state.create_job(command="sleep 1")
    j6 = state.create_job(command="sleep 1")


    j2.status = Job.Status.RUNNING
    j2.save()
    j3.status = Job.Status.RUNNING
    j3.save()
    j4.status = Job.Status.FAILED
    j4.save()
    j5.status = Job.Status.COMPLETED
    j5.save()
    j6.status = Job.Status.UNKNOWN
    j6.save()

    exp = {
        Job.Status.RUNNING: 2,
        Job.Status.SUBMITTED: 0,
        Job.Status.FAILED: 1,
        Job.Status.COMPLETED: 1,
        Job.Status.UNKNOWN: 1,
        Job.Status.CREATED: 1,
    }

    assert exp == f1.job_stats()
    assert exp == root.job_stats()



import os
import shutil
import stat
import subprocess
from concurrent.futures.thread import ThreadPoolExecutor
from datetime import timedelta
from io import StringIO
from pathlib import Path
from unittest.mock import Mock, ANY

import click
import pytest

from kong.util import (
    strip_colors,
    ljust,
    rjust,
    format_timedelta,
    parse_timedelta,
    shorten,
    shorten_path,
    make_executable,
    is_executable,
    rmtree,
    chunks,
    Spinner,
    Progress,
    get_size,
)


def test_strip_colors():
    s = "abcdef"
    assert strip_colors(s) == s

    assert strip_colors(click.style(s, fg="red")) == s


def test_ljust():
    s = "abcdlkjhdlfkjh"
    assert ljust(s, 30) == s.ljust(30)
    s_style = click.style(s, fg="red")
    assert ljust(s_style, 30)[-(30 - len(s)) :] == " " * (30 - len(s))


def test_rjust():
    s = "abcdlkjhdlfkjh"
    assert rjust(s, 30) == s.rjust(30)
    s_style = click.style(s, fg="red")
    assert rjust(s_style, 30)[: (30 - len(s))] == " " * (30 - len(s))


def test_format_timedelta():
    fmt = format_timedelta
    assert fmt(timedelta(seconds=30)) == "00:00:30"
    assert fmt(timedelta(minutes=42)) == "00:42:00"
    assert fmt(timedelta(minutes=42, seconds=14)) == "00:42:14"
    assert fmt(timedelta(hours=8)) == "08:00:00"
    assert fmt(timedelta(hours=99)) == "99:00:00"
    assert fmt(timedelta(days=3)) == f"{3*24}:00:00"
    with pytest.raises(ValueError):
        fmt(timedelta(days=5))
    assert fmt(timedelta(hours=99, minutes=59, seconds=59)) == "99:59:59"
    with pytest.raises(ValueError):
        fmt(timedelta(hours=100))
    assert fmt(timedelta(hours=8, minutes=6)) == "08:06:00"
    assert fmt(timedelta(hours=8, minutes=6, seconds=23)) == "08:06:23"


def test_parse_timedelta():
    ptd = parse_timedelta
    assert ptd("00:00:30") == timedelta(seconds=30)
    assert ptd("00:42:00") == timedelta(minutes=42)
    assert ptd("00:42:14") == timedelta(minutes=42, seconds=14)
    assert ptd("08:00:00") == timedelta(hours=8)
    assert ptd("99:00:00") == timedelta(hours=99)
    assert ptd(f"{3 * 24}:00:00") == timedelta(days=3)
    with pytest.raises(ValueError):
        ptd("100:00:00")
    assert ptd("99:59:59") == timedelta(hours=99, minutes=59, seconds=59)
    assert ptd("08:06:00") == timedelta(hours=8, minutes=6)
    assert ptd("08:06:23") == timedelta(hours=8, minutes=6, seconds=23)


def test_shorten():
    assert shorten("abcabcabcabc", 5) == "a...c"
    assert shorten("abcabcabcabc", 6) == "a...bc"
    assert shorten("abcabcabcabc", 7) == "ab...bc"
    assert shorten("abcdefghijklmnopqrstuvwxyz", 8) == "ab...xyz"
    assert shorten("abcdefghijklmnopqrstuvwxyz", 9) == "abc...xyz"
    assert shorten("abcabcabcabc", 12) == "abcabcabcabc"
    assert shorten("abcabcabcabc", 15) == "abcabcabcabc"
    with pytest.raises(ValueError):
        shorten("abcdefgh", 4)  # doesn't make sense


    s = click.style("abcabcabcabc", bold=True)
    out = shorten(s, 6)
    assert len(click.unstyle(out)) == 6
    assert click.unstyle(out) == "a...bc"
    assert out != click.unstyle(out)

    # mixed, will be stripped
    s = click.style("abcabcabcab", bold=True) + "c"
    out = shorten(s, 6)
    assert len(click.unstyle(out)) == 6
    assert click.unstyle(out) == "a...bc"
    assert out == click.unstyle(out)



def test_shorten_path():
    assert (
        shorten_path("/a/very/long/path/with/many/segments") == "/a/v/l/p/w/m/segments"
    )
    assert shorten_path("a/very/long/path/with/many/segments") == "a/v/l/p/w/m/segments"

    assert (
        shorten_path("a/very/long/path/with/many/segments_is_very_long_too", 10)
        == "a/v/l/p/w/m/seg..._too"
    )


def test_make_executable(tmp_path):
    p = tmp_path.joinpath("test.sh")
    p.write_text("hallo")

    mode = os.stat(p).st_mode
    assert (mode & stat.S_IEXEC) == 0

    make_executable(p)
    mode = os.stat(p).st_mode
    assert (mode & stat.S_IEXEC) != 0


def test_is_executable(tmp_path):
    p = tmp_path.joinpath("test.sh")
    p.write_text("hallo")
    assert is_executable(p) == False

    os.chmod(p, os.stat(p).st_mode | stat.S_IEXEC)
    assert is_executable(p) == True


def test_rmtree(monkeypatch):
    with monkeypatch.context() as m:
        rmtree_mock = Mock(side_effect=OSError)
        m.setattr("shutil.rmtree", rmtree_mock)
        system = Mock()
        m.setattr("os.system", system)
        rmtree("whatever/blaaa")
        rmtree_mock.assert_called_once_with("whatever/blaaa")
        system.assert_called_once_with("rm -rf whatever/blaaa")

    with monkeypatch.context() as m:
        rmtree_mock = Mock()
        m.setattr("shutil.rmtree", rmtree_mock)
        system = Mock()
        m.setattr("os.system", system)
        rmtree("whatever/blaaa")
        rmtree_mock.assert_called_once_with("whatever/blaaa")
        assert system.call_count == 0


def test_chunks():
    l = [1, 2, 3, 4, 5, 6, 7]
    ch = list(chunks(l, 2))

    assert ch == [[1, 2], [3, 4], [5, 6], [7]]


def test_spinner(monkeypatch):

    with monkeypatch.context() as m:
        write = Mock()
        isatty = Mock(return_value=False)
        m.setattr("sys.stdout.isatty", isatty)
        m.setattr("sys.stdout.write", write)
        with Spinner(text="blub"):
            pass
        write.assert_called_once_with("blub\n")

    with monkeypatch.context() as m:
        HaloInstance = Mock()
        HaloInstance.start = Mock()
        HaloInstance.succeed = Mock()
        HaloInstance.fail = Mock()
        HaloInstance.stop = Mock()
        Halo = Mock(return_value=HaloInstance)

        isatty = Mock(return_value=True)
        m.setattr("sys.stdout.isatty", isatty)
        m.setattr("kong.util.Halo", Halo)

        with Spinner(text="blub"):
            pass

        assert HaloInstance.start.call_count == 1
        assert HaloInstance.succeed.call_count == 1
        assert HaloInstance.stop.call_count == 0
        assert HaloInstance.fail.call_count == 0
        Halo.assert_called_once()

    with monkeypatch.context() as m:
        HaloInstance = Mock()
        HaloInstance.start = Mock()
        HaloInstance.succeed = Mock()
        HaloInstance.fail = Mock()
        HaloInstance.stop = Mock()
        Halo = Mock(return_value=HaloInstance)

        isatty = Mock(return_value=True)
        m.setattr("sys.stdout.isatty", isatty)
        m.setattr("kong.util.Halo", Halo)

        with pytest.raises(RuntimeError):
            with Spinner(text="blub", spinner="dots"):
                raise RuntimeError()

        assert HaloInstance.start.call_count == 1
        assert HaloInstance.succeed.call_count == 0
        assert HaloInstance.stop.call_count == 0
        assert HaloInstance.fail.call_count == 1
        Halo.assert_called_once_with("blub", spinner="dots")

    with monkeypatch.context() as m:
        HaloInstance = Mock()
        HaloInstance.start = Mock()
        HaloInstance.succeed = Mock()
        HaloInstance.fail = Mock()
        HaloInstance.stop = Mock()
        Halo = Mock(return_value=HaloInstance)

        isatty = Mock(return_value=True)
        m.setattr("sys.stdout.isatty", isatty)
        m.setattr("kong.util.Halo", Halo)

        with pytest.raises(RuntimeError):
            with Spinner(text="blub", persist=False):
                raise RuntimeError()

        assert HaloInstance.start.call_count == 1
        assert HaloInstance.succeed.call_count == 0
        assert HaloInstance.stop.call_count == 1
        assert HaloInstance.fail.call_count == 0

    with monkeypatch.context() as m:
        HaloInstance = Mock()
        HaloInstance.start = Mock()
        HaloInstance.succeed = Mock()
        HaloInstance.fail = Mock()
        HaloInstance.stop = Mock()
        Halo = Mock(return_value=HaloInstance)

        isatty = Mock(return_value=True)
        m.setattr("sys.stdout.isatty", isatty)
        m.setattr("kong.util.Halo", Halo)

        with Spinner(text="blub", persist=False):
            pass

        assert HaloInstance.start.call_count == 1
        assert HaloInstance.succeed.call_count == 0
        assert HaloInstance.stop.call_count == 1
        assert HaloInstance.fail.call_count == 0


def test_progress(monkeypatch):
    tqdm = Mock(return_value=[])
    monkeypatch.setattr("kong.util.tqdm", tqdm)
    for i in Progress(range(10)):
        pass
    assert tqdm.call_count == 1


@pytest.yield_fixture
def cleaned_tmpdir(tmpdir):
    sub = Path(tmpdir) / "subdir"
    sub.mkdir(exist_ok=True)
    shutil.rmtree(sub)
    sub.mkdir()
    yield sub


def test_get_size(cleaned_tmpdir):
    tmpdir = cleaned_tmpdir

    A = tmpdir / "A"
    B = tmpdir / "B"
    C = B / "C"

    A.mkdir()
    B.mkdir()
    C.mkdir()

    files = (
        [A / f"file_{i}" for i in range(4)]
        + [B / f"file_{i}" for i in range(8)]
        + [C / f"file_{i}" for i in range(2)]
    )

    for file in files:
        with file.open("wb") as fh:
            fh.write(b"X" * 1024)

    size = sum(os.path.getsize(f) for f in files)

    assert get_size(tmpdir) == size
    with ThreadPoolExecutor() as ex:
        assert get_size(tmpdir, ex) == size



"""
Singleton database instance
"""
from typing import TYPE_CHECKING, Any, List, ContextManager, Tuple, Iterable

if not TYPE_CHECKING:
    from playhouse.sqlite_ext import SqliteExtDatabase, AutoIncrementField
else:  # pragma: no cover

    class SqliteExtDatabase:
        """
        Mypy stub for the not type-hinted SqliteExtDatabase class
        """

        def __init__(self, *args: Any) -> None:
            """
            Type stub
            :param args:
            """
            ...

        def init(self, *args: Any) -> None:
            """
            Type stub
            :param args:
            :return:
            """
            ...

        def connect(self) -> None:
            """
            Type stub
            :return:
            """
            ...

        def create_tables(self, tables: List[Any]) -> None:
            """
            Type stub
            :param tables:
            :return:
            """
            ...

        def atomic(self) -> ContextManager[None]:
            """
            Type stub
            :return:
            """
            ...

        def execute_sql(self, query: str, params: Tuple[Any]) -> Iterable[Tuple[Any]]:
            ...

    class AutoIncrementField:
        """
        Type stub
        """

        ...


database = SqliteExtDatabase(None)



import os
import datetime

from typing import Any, cast, Optional, TYPE_CHECKING, List, Iterable, Dict, Tuple

import peewee as pw
from peewee import sqlite3

from ..logger import logger
from ..db import AutoIncrementField, database
from . import BaseModel

if TYPE_CHECKING:  # pragma: no cover
    from .job import Job


class Folder(BaseModel):
    """
    Represents a folder in the internal hierarchy for organizing jobs.

    :ivar folder_id: The ID of a folder instance
    :ivar name: Name of this folder instance
    :ivar parent: Points to the parent instance of this folder. Can be `None`
                  for the root folder
    :ivar created_at: Timestamp of creation of this folder instance
    :ivar updated_at: When the instance was last updated.


    """

    if TYPE_CHECKING:  # pragma: no cover
        folder_id: int
        parent: "Folder"
        children: List["Folder"]
        name: str
        jobs: List[Job]
    else:
        folder_id = AutoIncrementField()
        name = pw.CharField()
        parent = pw.ForeignKeyField("self", null=True, backref="children")
        created_at = pw.DateTimeField(default=datetime.datetime.now)
        updated_at = pw.DateTimeField()

    _ignore_save_assert = False

    class Meta:
        indexes = ((("parent", "name"), True),)

    def save(self, *args: Any, **kwargs: Any) -> None:
        if not self._ignore_save_assert:
            assert (
                self.name not in (".", "..", "")
                and "/" not in self.name
                and not self.name.isdigit()
            ), f"Invalid folder name '{self.name}'"
            assert self.parent is not None, "Need to specify a parent folder"
        self._ignore_save_assert = False

        # can never be its own parent
        if self.parent == self:
            raise pw.IntegrityError("Folder can not be its own parent")

        self.updated_at = datetime.datetime.now()

        return super().save(*args, **kwargs)

    @property
    def path(self) -> str:
        """
        The path to this folder

        :return: The path
        """
        if self.parent is None:
            return "/"
        # this will be slow, could optimize with CTE
        return cast(str, os.path.join(self.parent.path, self.name))

    @classmethod
    def get_root(cls) -> "Folder":
        """
        Retrieve the root folder (/). There can be only one.

        :return: Root folder instance
        """
        folder = Folder.get_or_none(Folder.parent.is_null(), name="root")
        if folder is None:
            # bypass assertions
            folder = super(BaseModel, cls).create(name="root", _ignore_save_assert=True)
        return cast(Folder, folder)

    @staticmethod
    def find_by_path(path: str, cwd: Optional["Folder"] = None) -> Optional["Folder"]:
        """
        Retrieve a folder instance by path

        :param path: Path to the folder
        :param cwd: Directory to start working from, defaults to root folder
        :return: The found folder or `None` if the path doesn't exist
        """

        assert isinstance(path, str)

        if cwd is None:
            cwd = Folder.get_root()

        assert isinstance(cwd, Folder)

        if path == "/":
            return Folder.get_root()
        if path.startswith("/"):
            return Folder.find_by_path(path[1:], Folder.get_root())
        if path.endswith("/"):
            path = path[:-1]
        if path == "..":
            return cwd.parent
        if path == "" or path == ".":
            return cwd
        if "/" not in path:
            return cwd.subfolder(path)
        head, tail = path.split(os.sep, 1)
        logger.debug("Resolve path %s in %s: %s, %s", path, cwd.path, head, tail)

        if head == "..":
            if cwd.parent is None:
                return None
            return Folder.find_by_path(tail, cwd.parent)
        else:
            next_cwd: Optional[Folder] = cwd.subfolder(head)
            if next_cwd is None:
                return None
            return Folder.find_by_path(tail, next_cwd)

    def __truediv__(self, name: str) -> Optional["Folder"]:
        return self.subfolder(name)

    def add_folder(self, name: str) -> "Folder":
        """
        Add a subfolder to this folder instance.

        :param name: Name of the new folder name (without /)
        :return: The new folder instance
        """
        return Folder.create(name=name, parent=self)

    def subfolder(self, name: str) -> Optional["Folder"]:
        """
        Retrieve a direct subfolder of this folder instance

        :param name: The unqualified name of the subfolder to retrieve.
        :return: Folder instance if `name` exists, else `None`
        """
        return Folder.get_or_none(Folder.parent == self, Folder.name == name)

    def folders_recursive(self) -> Iterable["Folder"]:
        """
        Recursively find all folders below this one.

        :return: All folders in the hierarchy from this folder. (Excludes this folder)
        """
        crit = (3, 8, 3)
        if sqlite3.sqlite_version_info < crit:  # pragma: no cover
            logger.debug(
                "sqlite3 version %s < %s: use slow python recursion",
                sqlite3.sqlite_version_info,
                crit,
            )
            folders: List["Folder"] = []
            for child in self.children:
                folders += child.folders_recursive()
            return folders

        else:
            logger.debug(
                "sqlite3 version %s >= %s: use CTE", sqlite3.sqlite_version_info, crit
            )
            sql = """
    WITH RECURSIVE
      children(n) AS (
        VALUES(?)
        UNION
        SELECT folder_id FROM folder, children
         WHERE folder.parent_id=children.n
      )
    SELECT * FROM folder where folder_id in children AND folder_id != ?;
    """

            return Folder.raw(sql, int(self.folder_id), int(self.folder_id))

    def jobs_recursive(self) -> Iterable["Job"]:
        """
        Recursively get all jobs in this folder and descendants.

        :return: Iterable over all jobs found, including jobs directly in this folder.
        """
        crit = (3, 8, 3)
        if sqlite3.sqlite_version_info < crit:  # pragma: no cover
            logger.debug(
                "sqlite3 version %s < %s: use slow python recursion",
                sqlite3.sqlite_version_info,
                crit,
            )
            jobs: List["Job"] = list(self.jobs)
            for child in self.children:
                jobs += child.jobs_recursive()
            return jobs
        else:
            logger.debug(
                "sqlite3 version %s >= %s: use CTE", sqlite3.sqlite_version_info, crit
            )
            sql = """
            WITH RECURSIVE
              children(n) AS (
                VALUES(?)
                UNION
                SELECT folder_id FROM folder, children
                 WHERE folder.parent_id=children.n
              )
            SELECT * FROM job where folder_id in children;
            """

            from .job import Job

            return Job.raw(sql, int(self.folder_id))

    def job_stats(self) -> Dict["Job.Status", int]:
        crit = (3, 8, 3)
        if sqlite3.sqlite_version_info < crit:  # pragma: no cover
            jobs = self.jobs_recursive()
            counts = {k: 0 for k in Job.Status}
            for job in jobs:
                counts[job.status] += 1
            return counts
        else:
            sql = """
WITH RECURSIVE
    children(n) AS (
       VALUES(?)
        UNION
        SELECT folder_id FROM folder, children
        WHERE folder.parent_id=children.n
    )
SELECT status, count() FROM job where folder_id in children GROUP BY status;
            """

            cursor = cast(
                Iterable[Tuple[int, int]],
                database.execute_sql(sql, (int(self.folder_id),)),
            )
            counts = {k: 0 for k in Job.Status}
            for status, count in cursor:
                counts[Job.Status(status)] = count

            return counts


# Needed for RTD
from .job import Job  # noqa: F402



from setuptools import setup, find_packages  # type: ignore
import os

thisdir = os.path.dirname(__file__)

with open(os.path.join(thisdir, "README.md")) as fh:
    long_desc = fh.read()

dev_requires = ["black"]
doc_requires = ["sphinx-autodoc-typehints"]
tests_require = [
    "pytest",
    "coverage",
    "pytest-cov",
    "pytest-rerunfailures",
    "mypy",
    "flake8",
    "tox"
]

def local_scheme(version):
    """Skip the local version (eg. +xyz of 0.6.1.dev4+gdf99fe2)
    to be able to upload to Test PyPI"""
    return ""

setup(
    name="kong-batch",
    use_scm_version={"local_scheme": local_scheme},
    setup_requires=["setuptools_scm"],
    description="Batch job submission and management tool",
    long_description=long_desc,
    long_description_content_type='text/markdown',
    url="http://github.com/paulgessinger/kong",
    author="Paul Gessinger",
    author_email="hello@paulgessinger.com",
    license="MIT",
    install_requires=[
        "click",
        "pyyaml",
        "halo",
        "sh",
        "python-dateutil",
        "peewee",
        "coloredlogs",
        "humanfriendly",
        "psutil",
        "schema",
        "tqdm",
        "notifiers",
        "jinja2",
        "pytimeparse",
    ],
    tests_require=tests_require,
    extras_require={
        "dev": dev_requires,
        "doc": doc_requires,
        "test": tests_require,
        "ipython": ["ipython"],
    },
    entry_points={"console_scripts": ["kong=kong.cli:main"]},
    packages=find_packages("src"),
    package_dir={"": "src"},
)
